#version 330 core
#auto_defines

#define PI     3.14159265
#define PI_2   1.5707963
#define invPI  0.318309886
#define FLUX_REL    9.167e-4
#define FLUX_MAX    0.1
#define edr  0.333333333  //  = 1/3
#define es   0.037037037  //  = 1/27
#define s3   1.732050808  //  sqrt(3)
#define s3h  0.866025404  //  sqrt(3/2)

//uniform sampler2D  texEllF;
uniform sampler1D  PlanckFunction;

uniform mat4x4  ModelView;
uniform mat4x4  Projection;
uniform vec3    RadDist; // Rg, dist, Rg/dist (in Rg units)
uniform vec2    DiscRadius;
uniform vec2    EmissionParams;
uniform float   iAngle;
uniform vec2    tfovXY;
uniform vec3    EyePos;
uniform vec3    DiscNorm;


#ifdef _VERTEX_

layout(location = 0) in  vec4  VertexPos;
layout(location = 1) in  vec4  VertexTexCoord;
                     out vec2  TexCoord;

void main()
{
    gl_Position = VertexPos;
	TexCoord = VertexTexCoord.xy;
}

#else

                     in  vec2  TexCoord;
layout(location = 0) out vec4  OutColor;

float a[15];
float b[15];
float c[15];
float phi[15];

float acosh(float x)
{
    return log(x + sqrt(x*x-1.0));
}

vec2 mulComplex(vec2 a, vec2 b)
{
    return vec2(a.x*b.x - a.y*b.y, a.x*b.y + a.y*b.x);
}

vec2 divComplex(vec2 a, vec2 b)
{
    float w = 1.0/(b.x*b.x + b.y*b.y);
    return w * vec2(a.x*b.x + a.y*b.y, a.y*b.x - a.x*b.y);
}

vec2 sqrtComplex(vec2 a)
{
    float r = sqrt(a.x*a.x + a.y*a.y);
    float f1 = 0.5*(r+a.x);
    float f2 = 0.5*(r-a.x);
    return vec2(sqrt(f1), sign(a.y)*sqrt(f2));
}

void transCartSph(in vec3 xOld, out vec3 xNew)
{
    float lenxy2 = xOld.x*xOld.x + xOld.y*xOld.y;
    xNew.x = sqrt(lenxy2 + xOld.z*xOld.z);
    xNew.y = atan(sqrt(lenxy2), xOld.z);
    xNew.z = atan(xOld.y, xOld.x);
}

void transCartSph(in vec3 oldPos, in vec3 oldDir, out vec3 newPos, out vec3 newDir)
{
    transCartSph(oldPos,newPos);
    newDir.x =  oldDir.x*sin(newPos.y)*cos(newPos.z)          + oldDir.y*sin(newPos.y)*sin(newPos.z)          + oldDir.z*cos(newPos.y);
    newDir.y =  oldDir.x*cos(newPos.y)*cos(newPos.z)/newPos.x + oldDir.y*cos(newPos.y)*sin(newPos.z)/newPos.x - oldDir.z*sin(newPos.y)/newPos.x;
    newDir.z = -oldDir.x*sin(newPos.z)/newPos.x/sin(newPos.y) + oldDir.y*cos(newPos.z)/newPos.x/sin(newPos.y);
}

// --------------------- sncndn function -------------------------

void sncndn(float u, float m, float eps, inout float sn, inout float cn, inout float dn)
{
    float mu,v,hn;
    if (m>1.0) { mu=1.0/m; v=u*sqrt(m); } else { mu=m; v=u; }
    a[0] = 1.0;
    b[0] = sqrt(1.0-mu);
    c[0] = sqrt(mu);
    int N=0;

    do
	{
        N++;
        a[N] = 0.5*(a[N-1]+b[N-1]);
        b[N] = sqrt(a[N-1]*b[N-1]);
        c[N] = 0.5*(a[N-1]-b[N-1]);
    }
	while (N < 14);

    phi[N] = pow(2.0,float(N))*a[N]*v;

    for(int i=N; i>0; i--)
        phi[i-1] = 0.5 * (asin(c[i]/a[i] * sin(phi[i])) + phi[i]);

    sn = sin(phi[0]);
    cn = cos(phi[0]);
    dn = cn/cos(phi[1]-phi[0]);
    if (m > 1.0)
	{
        sn *= sqrt(mu);
        hn = dn; dn = cn; cn = hn;
    }
}

// --------------------- einfach function -------------------------

float einfach(const float q, const float rho, float Phi1)
{
    float xi   = RadDist.z;
    float psi  = acos(q/pow(rho,3.0));
    float psi3 = psi*edr;
    float sq3  = sqrt(3.0);
    float cpsi3 = cos(psi3);
    float spsi3 = sin(psi3);
    float x1,x2,x3;

    if (rho > 0.0)
	{
        x1 = rho*(cpsi3-sq3*spsi3)+edr;
        x2 = rho*(cpsi3+sq3*spsi3)+edr;
        x3 = -2.0*rho*cpsi3+edr;
    }
	else
	{
        x1 = rho*(cpsi3-sq3*spsi3)+edr;
        x3 = rho*(cpsi3+sq3*spsi3)+edr;
        x2 = -2.0*rho*cpsi3+edr;
    }

    float x2mx1 = x2-x1;
    float x1mx3 = x1-x3;
    float x2mx3 = x2-x3;
    float kqua = x1mx3/x2mx3;
    float ximx1 = xi-x1;
    float ximx2 = xi-x2;
    float ximx3 = xi-x3;
    float u = 0.5*sqrt(x2mx3)*Phi1;
    float sn,cn,dn;

    sncndn(u,kqua,1.0e-12,sn,cn,dn);

    float numerator   = sn*sqrt(x2mx1*x2mx1 * ximx3 / (ximx1*ximx1*x2mx3)) + sqrt(ximx2/ximx1)*cn*dn;
    float denominator = 1.0 - x1mx3/x2mx3 * ximx2/ximx1*sn*sn;
    float snuv = numerator/denominator;
    return (x2-x1*snuv*snuv)/(1.0-snuv*snuv);
}

// --------------------- aufwendig function -------------------------

float aufwendig(const float q, const float rho, float Phi1)
{
    float xi   = RadDist.z;
    float psi  = acosh(sign(q)*abs(q)*27.0);
    float psi3 = psi*edr;
    float sq3  = sqrt(3.0);
    float esdr = 1.0/sq3;
	float expp = exp(psi3);
	float expm = exp(-psi3);
    float cpsi3 = (expp + expm) * 0.5;
    float spsi3 = (expp - expm) * 0.5;
    vec2 x1 = vec2(edr*(cpsi3+1.0), esdr*spsi3);
    vec2 x2 = vec2(-2.0*edr*cpsi3+edr, 0);
    vec2 x3 = vec2(edr*(cpsi3+1.0), -esdr*spsi3);
    vec2 kq = divComplex(x1-x3,x2-x3);
    vec2 ks = sqrtComplex(vec2(1.0,0.0)-kq);
    vec2 k1 = divComplex(vec2(1.0,0.0)-ks,vec2(1.0,0.0)+ks);
    float mchi = spsi3/cpsi3*esdr;
    float k2   = mchi/(1.0+sqrt(1.0+mchi*mchi));
    float zeta = sqrt(1.0+k2*k2);
    vec2 x_i = vec2(xi,0.0);
    vec2 cnvm = sqrtComplex(divComplex(x2-x3,x_i-x1));
    vec2 dnvm = sqrtComplex(divComplex(x_i-x3,x_i-x1));
    vec2 f1 = cnvm + dnvm + vec2(1.0,0.0) + mulComplex(ks,cnvm);
    vec2 f2 = cnvm + dnvm - vec2(1.0,0.0) - mulComplex(ks,cnvm);
    vec2 cnz0c = mulComplex(k1,divComplex(f1,f2));
    float ezeta2 = 1.0/(zeta*zeta);
    float cnz0 = cnz0c.x;
    float snz0 = sqrt(1.0-cnz0*cnz0);
    float dnz0 = sqrt(1.0-snz0*snz0*ezeta2);
    float z = 0.5*Phi1*sqrt(cpsi3)*pow(1.0+mchi*mchi,0.25);
    float sn,cn,dn;
    sncndn(z,ezeta2,1.0e-12,sn,cn,dn);
    float SD = (sn*cnz0*dnz0 + snz0*cn*dn)/(dn*dnz0 - ezeta2*sn*snz0*cn*cnz0);
    float SD2 = pow(SD,2.0);
    float SD4 = pow(SD,4.0);
    float numerator   = (1.0-2.0*cpsi3)*(1.0-pow(k2*ezeta2,2.0)*SD4) - ((cpsi3+1.0)*(k2*k2-1.0)-2.0*sq3*k2*spsi3)*ezeta2*SD2;
    float denominator = 3.0*(1.0-pow(k2*ezeta2,2.0)*SD4 + (1.0-k2*k2)*ezeta2*SD2);
    return numerator/denominator;
}

// --------------------- norm function -------------------------

float norm(vec4 pos, vec4 v, vec4 w)
{
    float r     = pos.x;
    float theta = pos.y;
    float t4 = -1.0 + RadDist.x / r;
    float t6 = r*r;
    float t7 = sin(theta);
    float t8 = t7*t7;
    float g_compts[4];
    g_compts[3] = t4;
    g_compts[0] = -1.0/t4;
    g_compts[1] = t6;
    g_compts[2] = t6*t8;
    float sum = 0.0;
    for(int i=0; i<4; i++) 
        sum += g_compts[i]*v[i]*w[i];
    return sum;
}

// --------------------- redshift function -------------------------

float redShift(float sinksi, vec3 hitPosSph, vec3 moveDirSph, float mchi, float angleQ, float omega)
{
    float mchi2  = mchi*mchi;
	float rsink  = RadDist.y * sinksi;
	float s1mdr  = sqrt(1.0-RadDist.z);
    float kdh    = s1mdr / rsink;
    float drdphi = abs(mchi2 * (mchi2 * kdh*kdh - 1.0 + RadDist.x/mchi));
    drdphi = sqrt(drdphi);
    float dtdphi = kdh * mchi2 / (1.0-RadDist.x/mchi);
    float phip   = rsink / mchi2;
    vec4  k      = vec4(phip*drdphi, 0.0, phip, phip*dtdphi);
    vec4  pos    = vec4(mchi, PI_2, angleQ, 0.0);

    // particle moves on circular timelike geodesic
    float v = 1.0 / sqrt(2.0*(mchi/RadDist.x-1.0));
    float gamma = 1.0 / sqrt(1.0-v*v);

    // particle's four-velocity
    vec4  u = gamma * vec4(v * vec3(s1mdr, 1.0, 1.0) * moveDirSph, 1.0 / s1mdr);
    float wEmit = norm(pos, k, u);
    float wObs = -1.0;
    return wEmit/wObs;
}

// --------------------- main shader function -------------------------

void main()
{
    for(int i=0; i<15; i++)
         a[i] = b[i] = c[i] = phi[1] = 0.0;

    // direction of geodesic
	//vec2  Ray = (1.0 - 2.0 * TexCoord) * tfovXY.xy;
	vec2 Plane = (1.0 - 2.0 * TexCoord) * tfovXY.xy;
	//vec4  Ray4 = vec4(normalize(vec3(Plane, 1.0)), 0.0);
	vec4 Ray4 = vec4(Plane, 1.0, 0.0);
	//Ray4 += vec4(EyePos, 0.0);
	//Ray4 = Ray4 * ModelView;
	vec2 Ray = Ray4.xy;

    // geodesic plane is rotated by omega out of the equatorial plane
    float omega = atan(Ray.y, Ray.x);

    float normxd, normdr, aqua, q, psi, rho;
    float ti  = tan(iAngle);
    float ti2 = ti*ti;
    float ci  = cos(iAngle);
    float si  = sin(iAngle);
    float vs  = 4.0*es;
    float angleQ;      // intersection angle
    float chi = 0.0;   // intersection (distance to black hole, azimuth)

    // initial direction of the geodesic within the geodesic plane
    normdr = sqrt(1.0 + dot(Ray, Ray));

    // initial angle
	float sinksi = sin(acos(1.0/normdr));

    // parameter for analytic solution
    aqua = RadDist.z*RadDist.z * (1.0-RadDist.z) / (sinksi * sinksi);
    q    = aqua*0.5-es;
    rho  = sign(q)*edr;

    // determine intersection angle between geodesic and disk
    normxd = sqrt(Ray.x*Ray.x + Ray.y*Ray.y * (1.0 + ti2));
    angleQ = acos(-Ray.y * ti / normxd);
    int side = 0;  // noside = 0, front = 1, back = 2

    if (aqua < vs)
	{
        chi = RadDist.x / einfach(q, rho, angleQ);
        if ((chi < DiscRadius.x) || (chi > DiscRadius.y))
		{
            angleQ += PI;
            chi = RadDist.x / einfach(q, rho, angleQ);
            if ((chi > DiscRadius.y) || (chi < RadDist.x))
			    chi = 0.0;
            else
			    side = 2;
        }
    }
    else
	{
        chi = RadDist.x / aufwendig(q, rho, angleQ);
        side = 1;
        if ((chi < DiscRadius.x) || (chi > DiscRadius.y))
		    side = 3;
    }

    float R = chi/RadDist.x;
    float sR = sqrt(R);
    float flux = FLUX_MAX/FLUX_REL / max(R - 1.5, 0.001) * pow(R, -2.5) * (sR - s3 + 0.5*s3h * log((sR+s3h)/(sR-s3h) * (s3-s3h)/(s3+s3h)));

    OutColor.rgb = vec3(0.0);
	OutColor.a = clamp(flux * 100.0 * (1.0 - smoothstep(0.95, 1.0, chi/DiscRadius.y)), 0.0, 1.0);

    vec3 hitPos, m, moveDir, hitPosSph, moveDirSph;

    if ((chi > DiscRadius.x) && (chi < DiscRadius.y))
	{
        float edn = 1.0/normxd;

        // hit position in geod.plane
        hitPos  = chi*vec3(cos(angleQ), sin(angleQ), 0.0);

        // direction of motion (global)
        m = edn * vec3(-Ray.x * si, -Ray.y * (si*ti+ci), Ray.x * ci);

        // direction of motion wrt. geod.plane
        moveDir = vec3(m.x, m.y*cos(omega) + m.z*sin(omega), -m.y*sin(omega) + m.z*cos(omega));

        if (side == 2)
            moveDir = -moveDir;

        transCartSph(hitPos, moveDir, hitPosSph, moveDirSph);

        // Doppler redshift
        float onePlusZ = redShift(sinksi, hitPosSph, moveDirSph, chi, angleQ, omega);

        // gravitational redshift
        onePlusZ *= sqrt((1.0-RadDist.z) / (1.0-RadDist.x/chi));

		float onePlusZ2 = onePlusZ * onePlusZ;
        flux /= onePlusZ2 * onePlusZ2;

        float tempApp = pow(flux * 5e15 / onePlusZ, 0.25);
        OutColor.rgb = texture(PlanckFunction, log(tempApp * 0.001 + EmissionParams.x) * 0.188 + 0.1316).rgb * EmissionParams.y;
        //OutColor.rgb = vec3(flux);
        //OutColor = vec4(vec3(onePlusZ-1.0), 1.0);
    }

    if (side == 3)
	    OutColor = vec4(0.0);
/**/
}

#endif
