#version 330 compatibility
#auto_defines

//attribute  vec3  gl_Vertex;           // sprite center world position
//attribute  vec4  gl_Color;            // phase (hi byte), phase (lo byte), sprite type, sprite id

uniform mat4x4    ProjectionMatrix;
uniform mat4x4    ModelViewMatrix;
uniform vec4      ScaleTime;    // (scale, time)
uniform vec3      GasColor;
uniform vec3      DustColor;
uniform vec3      EyePos;
uniform vec3      CometPos;
uniform vec3      CometVel;
uniform vec3      Light0Pos;

#ifdef _VERTEX_

out vec3  FragPos;
out vec4  TexCoord;
out vec4  Color;
out float SpriteSize;

void main()
{
    TexCoord.x = mod(floor(gl_Color.a * 4.0), 2.0);
    TexCoord.y = mod(floor(gl_Color.a * 2.0), 2.0);

    float dist     = length(gl_Vertex.xyz);
    vec3  posVec   = gl_Vertex.xyz / dist;
    vec3  sunVec   = CometPos - Light0Pos;
    float sunDist  = length(sunVec);
    sunVec = sunVec / sunDist;
    float sunInvDist2 = min(2.25e16 / (sunDist * sunDist), 1.0);

    //float t = fract(ScaleTime.y * (1.0 + gl_Color.b) + gl_Color.r + gl_Color.g / 256.0);
    float t = fract(ScaleTime.y + gl_Color.r + gl_Color.g / 256.0);
    vec3  pos, vel, acc, spriteDir;
    float fade, scale, spriteSize, spriteLength;

    if (gl_Color.b == 0.0)
    {
        // gas particles - interaction with solar wind
        Color = vec4(GasColor, 0.0);
        TexCoord.w = 0.0;
        vel        =  0.03 * (1.000 + 13.0 * ScaleTime.z) * posVec;
        vec3  wvel = -14.0 * (1.001 -        ScaleTime.z) * sunVec;
        vec3  dv = dot(vel, sunVec) * sunVec - wvel;
        acc = 0.005 * dot(dv, dv) * sunVec;
        pos = posVec * 0.001 + vel * t + acc * t*t;

        fade  = (1.0 - pow(t, 0.25)) * clamp(t * 50.0, 0.0, 1.0) * clamp(1.0 - ScaleTime.z, 0.0, 1.0) / ScaleTime.w;
        scale = 1.0 + 10.0 * t + 10.0 * ScaleTime.z;
        spriteSize = ScaleTime.x * scale * 0.0025 * ScaleTime.w;
        spriteLength = 0.0001 + 0.1 * pow(clamp(t * 2.0, 0.0, 1.0), 4.0);
        spriteDir = sunVec * spriteLength;
    }
    else
    {
        // dust particles - interaction with solar light (light pressure)
        Color = vec4(DustColor, 0.0);
        TexCoord.w = 1.0;
        vel = 0.1 * (1.000 + 13.0 * ScaleTime.z) * posVec;
        acc = 0.5 * sunInvDist2 * sunVec - 0.02 * gl_Color.b * CometVel;
        //vel = 0.03 * (1.000 + 13.0 * ScaleTime.z) * posVec - 0.0003 * CometVel + 0.2 * gl_Color.b * sunVec;
        //acc = -0.02 * sunInvDist2 * sunVec;
        pos = posVec * 0.001 + vel * t + acc * t*t;

        fade  = (1.0 - pow(t, 0.25)) * clamp(t * 50.0, 0.0, 1.0) * clamp(1.0 - ScaleTime.z, 0.0, 1.0) / ScaleTime.w;
        scale = 1.0 + 10.0 * t + 10.0 * ScaleTime.z;
        spriteSize = ScaleTime.x * scale * 0.0075 * ScaleTime.w;
        spriteLength = 0.0001 + 0.1 * pow(clamp(t * 2, 0.0, 1.0), 4.0);
        spriteDir = normalize(acc) * spriteLength;
    }

    //vec3  eyeVec  = posos * ScaleTime.x + CometPos - EyePos;
    //float eyeDist = length(eyeVec);
    //spriteSize *= clamp(2.0 * ScaleTime.w / ScaleTime.x * eyeDist, 0.25, 1.0);

    vec4  viewPosN = ModelViewMatrix * vec4(pos, 1.0);
    vec4  viewPosO = ModelViewMatrix * vec4(pos + spriteDir, 1.0);

    vec4  viewPos;
    if (TexCoord.x == 0.0)
    {
        viewPos = viewPosN;
        FragPos = pos;
    }
    else
    {
        viewPos = viewPosO;
        FragPos = pos + spriteDir;
    }

    vec2  Offset = 1.0 - TexCoord.xy * 2.0;
    vec3  Ray    = viewPosN.xyz - viewPosO.xyz;
    float RayLen = length(viewPosN.xy - viewPosO.xy);

    float l = RayLen / spriteSize;

    /*if (l <= 1.001)
    {
        viewPos = viewPosN;
        viewPos.xy += Offset * spriteSize;
        TexCoord.z = 0.0;
    }
    else*/
    {
        /*float lmax = 5.0;
        if (l <= lmax)
        {
            float d = (l - 1.001) / (lmax - 1.001);
            vec4 viewPosM = (viewPosN + viewPosO) * 0.5;
            viewPos = mix(viewPosM, viewPos, d);
        }*/
        RayLen = length(viewPosN.xy - viewPosO.xy);
        TexCoord.z = RayLen / (RayLen + 2.0*spriteSize);
        vec3 Tangent = normalize(cross(Ray, viewPosN.xyz));
        mat2x2 Mat = mat2x2(Tangent.y, -Tangent.x, Tangent.x, Tangent.y);
        viewPos.xy += (Mat * Offset) * spriteSize;
        //FragPos.xy += (Mat * Offset) * spriteSize;
    }

    gl_Position = ProjectionMatrix * viewPos;

    FragPos = FragPos * ScaleTime.x / 149598000.0 + CometPos;
    SpriteSize = 3.0 * spriteSize / 149598000.0;
    Color *= fade;
}

#else

in vec3  FragPos;
in vec4  TexCoord;
in vec4  Color;
in float SpriteSize;

float HG(float cosa) // Henyey-Greenstein phase function
{
    const float g = 0.3;
    const float maxcosa = 0.4999 + 0.5*g*g;
    //float x = cosa - 0.97;
    //float halo = exp(-x*x/0.000024);
    return 0.5 * (1.0 - g*g) / pow(1.0 + g*g - 2.0*g*min(cosa, maxcosa), 1.5);// + 3.0 * halo;
}

void main()
{
    vec2 tc = TexCoord.xy * 2.0 - 1.0;
    if (tc.x < 0)
        tc.x = max((tc.x + TexCoord.z) / (TexCoord.z - 1.0), 0.0);
    else
        tc.x = min((tc.x - TexCoord.z) / (TexCoord.z - 1.0), 0.0);

    float  r = length(tc);
    float  bright = exp(-1.5 * r) * clamp(1.0f-r, 0.0, 1.0);

    vec3  eyePos    = (EyePos    - FragPos);
    vec3  light0Pos = (Light0Pos - FragPos);
    bright *= clamp(length(eyePos) / SpriteSize, 0.0, 1.0);
    bright *= gl_LightSource[0].diffuse.a / (dot(light0Pos, light0Pos) + 0.05);
    bright = pow(bright, 0.35);

    if (TexCoord.w == 0.0)
    {
        gl_FragColor = Color * bright;
    }
    else
    {
        float cosa = -dot(normalize(eyePos), normalize(light0Pos));
        gl_FragColor = Color * gl_LightSource[0].diffuse * (bright * bright / gl_LightSource[0].diffuse.a * HG(cosa));
    }
}

#endif
